/* $Id: isaac_img_twflat.c,v 1.45 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.45 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"
#include "irplib_flat.h"

#include "isaac_utils.h"
#include "irplib_oddeven.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/* FIXME: To be removed when CPL4.2 is available */
#define PRO_DATANCOM                "ESO PRO DATANCOM"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_img_twflat_create(cpl_plugin *);
static int isaac_img_twflat_exec(cpl_plugin *);
static int isaac_img_twflat_destroy(cpl_plugin *);
static int isaac_img_twflat(cpl_parameterlist *, cpl_frameset *);
static cpl_imagelist * isaac_img_twflat_reduce(cpl_frameset *, cpl_frameset *);
static int isaac_img_twflat_save(cpl_imagelist *, cpl_image *, int, 
                                 const cpl_frameset *, cpl_parameterlist *,
                                 cpl_frameset *);
static int isaac_img_twflat_compare(const cpl_frame *, const cpl_frame *); 

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    double      low_thresh;
    double      high_thresh;
    int         nb_badpix;
    double      med_stdev;
    double      med_avg;
    double      med_min;
    double      med_max;
    int         prop_flag;
    int         bpm_flag;
    int         errmap_flag;
    int         intercept_flag;
    double      oddeven_ll_max;
    double      oddeven_lr_max;
    double      oddeven_ul_max;
    double      oddeven_ur_max;
    double      oddeven_ll_stdev;
    double      oddeven_lr_stdev;
    double      oddeven_ul_stdev;
    double      oddeven_ur_stdev;
    double      oddeven_ll_mean;
    double      oddeven_lr_mean;
    double      oddeven_ul_mean;
    double      oddeven_ur_mean;
} isaac_img_twflat_config;

static char isaac_img_twflat_description[] = 
"isaac_img_twflat -- ISAAC imaging flat-field creation from twillight images.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "ISAAC_IMG_TWFLAT_RAW" or\n"
"raw-file.fits "ISAAC_IMG_TWFLAT_POL_RAW" or\n"
"dark-raw-file.fits "ISAAC_CALIB_DARK"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_img_twflat",
                    "Twillight flat recipe",
                    isaac_img_twflat_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_img_twflat_create,
                    isaac_img_twflat_exec,
                    isaac_img_twflat_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_twflat_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --t */
    p = cpl_parameter_new_value("isaac.isaac_img_twflat.thresholds",
                                CPL_TYPE_STRING,
                                "Low and high thresholds for the BPM",
                                "isaac.isaac_img_twflat",
                                "0.5,2.0");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "t");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --prop */
    p = cpl_parameter_new_value("isaac.isaac_img_twflat.proport", CPL_TYPE_BOOL,
            "flag to activate the proportional fit", "isaac.isaac_img_twflat",
            FALSE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "prop");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    
    /* --bpm */
    p = cpl_parameter_new_value("isaac.isaac_img_twflat.bpm", CPL_TYPE_BOOL,
            "flag to create the bad pixels map", "isaac.isaac_img_twflat",
            TRUE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "bpm");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    
    /* --errmap */
    p = cpl_parameter_new_value("isaac.isaac_img_twflat.errmap", CPL_TYPE_BOOL,
            "flag to create the error map", "isaac.isaac_img_twflat",
            FALSE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "errmap");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    
    /* --intercept */
    p = cpl_parameter_new_value("isaac.isaac_img_twflat.intercept", 
            CPL_TYPE_BOOL, "flag to create the intercept image", 
            "isaac.isaac_img_twflat", FALSE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "intercept");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    
    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_twflat_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_img_twflat(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_twflat_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_twflat(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    const char      *   sval;
    cpl_parameter   *   par;
    cpl_size        *   labels;
    cpl_size            nlabels;
    cpl_frameset    *   flatframes;
    cpl_frameset    *   darkframes;
    cpl_frameset    *   flat_one;
    cpl_frameset    *   dark_one;
    int                 ndarks;
    cpl_imagelist   *   twflat;
    cpl_mask        *   bpm;
    cpl_image       *   bpm_im;
    cpl_boolean         did_reduce = CPL_FALSE;
    int                 i;
    
    /* Initialise */
    par = NULL;
    bpm_im = NULL;
    isaac_img_twflat_config.nb_badpix = -1;

    /* Retrieve input parameters */
    /* --t */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_twflat.thresholds");
    sval = cpl_parameter_get_string(par);
    if (sscanf(sval, "%lg,%lg",
                    &isaac_img_twflat_config.low_thresh,
                    &isaac_img_twflat_config.high_thresh)!=2) {
        return CPL_ERROR_UNSPECIFIED;
    }
    /* --prop */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_twflat.proport");
    isaac_img_twflat_config.prop_flag = cpl_parameter_get_bool(par);
    /* --bpm */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_twflat.bpm");
    isaac_img_twflat_config.bpm_flag = cpl_parameter_get_bool(par);
    /* --errmap */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_twflat.errmap");
    isaac_img_twflat_config.errmap_flag = cpl_parameter_get_bool(par);
    /* --intercept */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_twflat.intercept");
    isaac_img_twflat_config.intercept_flag = cpl_parameter_get_bool(par);
   
    /* Identify the RAW and CALIB frames in the input frameset */
    if (isaac_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Retrieve raw frames */
    if ((flatframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_TWFLAT_RAW)) != NULL) {
    } else if ((flatframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_TWFLAT_POL_RAW)) != NULL) {
    } else {
        cpl_msg_error(cpl_func, "Cannot find flat frames in the input list");
        return CPL_ERROR_UNSPECIFIED;
    }
    darkframes = isaac_extract_frameset(framelist, ISAAC_CALIB_DARK);

    /* If dark frames are provided - check their number  */
    if (darkframes != NULL) {
        ndarks = cpl_frameset_get_size(darkframes);
        if ((ndarks != 1) && (cpl_frameset_get_size(flatframes) != ndarks)) {
            cpl_msg_error(cpl_func, "Invalid number of dark frames submitted");
            cpl_frameset_delete(flatframes);
            cpl_frameset_delete(darkframes);
            return CPL_ERROR_UNSPECIFIED;
        }
    }
    
    /* Labelise all input frames */
    labels = cpl_frameset_labelise(flatframes, isaac_img_twflat_compare, 
                &nlabels);
    if (labels == NULL) {
        cpl_msg_error(cpl_func, "Cannot labelise input frames");
        cpl_frameset_delete(flatframes);
        if (darkframes) cpl_frameset_delete(darkframes);
        return CPL_ERROR_UNSPECIFIED;
    }
   
    /* Extract settings and reduce each of them */
    for (i=0; i<nlabels; i++) {
        /* Reduce data set nb i */
        cpl_msg_info(cpl_func, "Reduce data set no %d out of %d", i+1,
                     (int)nlabels);
        cpl_msg_indent_more();
        flat_one = cpl_frameset_extract(flatframes, labels, i);
        dark_one = NULL;
        if (darkframes) dark_one = cpl_frameset_extract(darkframes, labels, i);
        twflat = isaac_img_twflat_reduce(flat_one, dark_one);
        cpl_msg_indent_less();

        /* Save the products */
        cpl_msg_info(cpl_func, "Save the products");
        cpl_msg_indent_more();
        if (twflat == NULL) {
            cpl_msg_warning(cpl_func, "Cannot reduce set nb %d", i+1);
        } else {
            /* Create the Bad pixels map */
            if (isaac_img_twflat_config.bpm_flag) {
                if ((bpm = cpl_mask_threshold_image_create(
                                cpl_imagelist_get(twflat, 0),
                                isaac_img_twflat_config.low_thresh,
                                isaac_img_twflat_config.high_thresh)) == NULL) {
                    cpl_msg_warning(cpl_func, "Cannot create bad pixels map");
                } else {
                    cpl_mask_not(bpm);
                    isaac_img_twflat_config.nb_badpix = cpl_mask_count(bpm);
                    bpm_im = cpl_image_new_from_mask(bpm);
                    cpl_mask_delete(bpm);
                }
            }
            isaac_img_twflat_save(twflat, bpm_im, i+1, flat_one, parlist, 
                    framelist);
            if (bpm_im) cpl_image_delete(bpm_im);
            cpl_imagelist_delete(twflat);
           if (!cpl_error_get_code()) did_reduce = CPL_TRUE;
        }
        cpl_msg_indent_less();
        cpl_frameset_delete(flat_one);
        if (dark_one) cpl_frameset_delete(dark_one);
    }
    
    /* Free and return */
    cpl_frameset_delete(flatframes);
    if (darkframes) cpl_frameset_delete(darkframes);
    cpl_free(labels); 

    cpl_ensure_code(did_reduce, CPL_ERROR_ILLEGAL_INPUT);

    return cpl_error_set_where(cpl_func); /* Propagate error, if any */
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    flatframes  the frames list with flat frames
  @param    darkframes  the frames list with dark frames
  @return   the image list with the results
  The first imageis the flat
  The second image is the intercept map in non-proport mode, and the
  error map otherwise.
  There is a third image in non-prop mode which is the err map.
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_img_twflat_reduce(
        cpl_frameset    *   flatframes,
        cpl_frameset    *   darkframes)
{
    cpl_frame           *   frame;
    cpl_propertylist    *   plist;
    const char          *   sval;
    cpl_image           *   flat_image;
    cpl_imagelist       *   in,
                        *   dark,
                        *   results;
    cpl_stats           *   stats_ima;
    cpl_vector          *   medians;
    double                  gradient;
    double                  norm;
    int                     nima;
    cpl_vector          *   reve_ll;
    cpl_vector          *   reve_lr;
    cpl_vector          *   reve_ul;
    cpl_vector          *   reve_ur;
    double                  ll, lr, ul, ur;
    int                     i;

    /* Test entries */
    if (flatframes == NULL) return NULL;

    /* Initialise */
    nima = cpl_frameset_get_size(flatframes);
    
    /* Check the error status */
    if (cpl_error_get_code()) return NULL;

    /* Print out the filter and the read-out mode */
    frame = cpl_frameset_get_position(flatframes, 0);
    plist=cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot get the filter or ROM or TPL.ID");
        cpl_propertylist_delete(plist);
        return NULL;
    }
    cpl_msg_info(cpl_func, "Filter:        [%s]", sval);
    cpl_propertylist_delete(plist);

    /* Load input image set */
    cpl_msg_info(cpl_func, "---> Loading input set");
    if ((in = cpl_imagelist_load_frameset(flatframes, CPL_TYPE_FLOAT, 1,
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load the image set");
        return NULL;
    }
    
    /* Compute some stats on input images */
    cpl_msg_info(cpl_func, "---> Computing stats");
    cpl_msg_info(cpl_func, "image      min        max        med     rms");
    cpl_msg_info(cpl_func, "---------------------------------------------");
    medians = cpl_vector_new(nima);
    for (i=0; i<nima; i++) {
        stats_ima = cpl_stats_new_from_image(cpl_imagelist_get(in, i), 
                CPL_STATS_ALL);
        if (stats_ima == NULL) {
            cpl_msg_error(cpl_func, "Cannot compute stats on image %d", i+1);
            cpl_imagelist_delete(in);
            cpl_vector_delete(medians);
            return NULL;
        }
        cpl_vector_set(medians, i, cpl_stats_get_median(stats_ima));
        cpl_msg_info(cpl_func, "%02d   %10.2f %10.2f %10.2f %10.2f",
                i+1,
                cpl_stats_get_min(stats_ima),
                cpl_stats_get_max(stats_ima),
                cpl_stats_get_median(stats_ima),
                cpl_stats_get_stdev(stats_ima));
        if (cpl_stats_get_median(stats_ima) < 1e-6) {
            cpl_msg_error(cpl_func, "image %d has negative flux: aborting", 
                    i+1);
            cpl_imagelist_delete(in);
            cpl_vector_delete(medians);
            cpl_stats_delete(stats_ima);
            return NULL;
        }
        cpl_stats_delete(stats_ima);
    }
    cpl_msg_info(cpl_func, "---------------------------------------------");

    /* Compute min max stdev and mean of the medians */
    isaac_img_twflat_config.med_min   = cpl_vector_get_min(medians);
    isaac_img_twflat_config.med_max   = cpl_vector_get_max(medians);
    isaac_img_twflat_config.med_avg   = cpl_vector_get_mean(medians);
    isaac_img_twflat_config.med_stdev = cpl_vector_get_stdev(medians);
    cpl_vector_delete(medians);
 
    /* Apply dark correction to all planes if requested */
    if (darkframes) {
        cpl_msg_info(cpl_func, "---> Subtracting dark");
        /* Load dark cube */
        if ((dark = cpl_imagelist_load_frameset(darkframes, CPL_TYPE_FLOAT, 1,
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the darks set");
            cpl_imagelist_delete(in);
            return NULL;
        }
        /* Dark correction */
        if (cpl_imagelist_get_size(dark) == 1) {
            cpl_imagelist_subtract_image(in, cpl_imagelist_get(dark, 0));
        } else {
            cpl_imagelist_subtract(in, dark);
        }
        cpl_imagelist_delete(dark);

        /* With the darks provided, the proportional method is used */
        cpl_msg_info(cpl_func, "switching to proportional fit");
        isaac_img_twflat_config.prop_flag = 1;
    }
    
    /* See if flux gradient is large enough for a correct fit */
    if (!isaac_img_twflat_config.prop_flag) {
        gradient=fabs(isaac_img_twflat_config.med_max/
                isaac_img_twflat_config.med_min);
        if (gradient < 4.0) {
            cpl_msg_warning(cpl_func, "low flux gradient: %g", gradient);
            cpl_msg_warning(cpl_func,"a proportional fit may give better ");
            cpl_msg_warning(cpl_func,"results (requires a master dark frame)");
        }
    }

    /* Monitor the odd/even column effect */
    cpl_msg_info(cpl_func, "---> Monitor odd/even column effect");
    reve_ll = cpl_vector_new(nima);
    reve_lr = cpl_vector_new(nima);
    reve_ul = cpl_vector_new(nima);
    reve_ur = cpl_vector_new(nima);
    for (i=0; i<nima; i++) {
        irplib_oddeven_monitor(cpl_imagelist_get(in, i), 1, &ll);
        irplib_oddeven_monitor(cpl_imagelist_get(in, i), 2, &lr);
        irplib_oddeven_monitor(cpl_imagelist_get(in, i), 3, &ul);
        irplib_oddeven_monitor(cpl_imagelist_get(in, i), 4, &ur);
        cpl_vector_set(reve_ll, i, ll);
        cpl_vector_set(reve_lr, i, lr);
        cpl_vector_set(reve_ul, i, ul);
        cpl_vector_set(reve_ur, i, ur);
    }
    isaac_img_twflat_config.oddeven_ll_max = cpl_vector_get_max(reve_ll);
    isaac_img_twflat_config.oddeven_lr_max = cpl_vector_get_max(reve_lr);
    isaac_img_twflat_config.oddeven_ul_max = cpl_vector_get_max(reve_ul);
    isaac_img_twflat_config.oddeven_ur_max = cpl_vector_get_max(reve_ur);
    isaac_img_twflat_config.oddeven_ll_stdev = cpl_vector_get_stdev(reve_ll);
    isaac_img_twflat_config.oddeven_lr_stdev = cpl_vector_get_stdev(reve_lr);
    isaac_img_twflat_config.oddeven_ul_stdev = cpl_vector_get_stdev(reve_ul);
    isaac_img_twflat_config.oddeven_ur_stdev = cpl_vector_get_stdev(reve_ur);
    isaac_img_twflat_config.oddeven_ll_mean = cpl_vector_get_mean(reve_ll);
    isaac_img_twflat_config.oddeven_lr_mean = cpl_vector_get_mean(reve_lr);
    isaac_img_twflat_config.oddeven_ul_mean = cpl_vector_get_mean(reve_ul);
    isaac_img_twflat_config.oddeven_ur_mean = cpl_vector_get_mean(reve_ur);
    cpl_vector_delete(reve_ll);
    cpl_vector_delete(reve_lr);
    cpl_vector_delete(reve_ul);
    cpl_vector_delete(reve_ur);

    /* Fit slopes, get results */
    cpl_msg_info(cpl_func, "---> Fitting slopes");
    if (isaac_img_twflat_config.prop_flag) {
        results = irplib_flat_fit_set(in, 0);
        if (results == NULL) {
            cpl_msg_error(cpl_func, "creating twilight flat-field: aborting");
            cpl_imagelist_delete(in);
            return NULL;
        }
    } else {
        results = irplib_flat_fit_set(in, 1);
        if (results == NULL) {
            cpl_msg_error(cpl_func, "creating twilight flat-field: aborting");
            cpl_imagelist_delete(in);
            return NULL;
        }
    }
    cpl_imagelist_delete(in);

    /* Normalize gain */
    flat_image = cpl_imagelist_get(results, 0);
    norm = cpl_image_get_mean(flat_image);
    cpl_image_divide_scalar(flat_image, norm);
    if (cpl_error_get_code()) {
        cpl_imagelist_delete(results);
        return NULL;
    }

    return results;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the lampflat recipe products on disk
  @param    flat    the flat images produced
  @param    set_nb  current setting number
  @param    set     the frames from the current setting
  @param    parlist     the input list of parameters
  @param    set_tot the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_twflat_save(
        cpl_imagelist       *   flat,
        cpl_image           *   bpm,
        int                     set_nb,
        const cpl_frameset  *   set,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set_tot)
{
    const cpl_size          nframes = cpl_frameset_get_size(set);
    cpl_propertylist    *   plist;
    cpl_propertylist    *   paflist;
    cpl_propertylist    *   qclist;
    const cpl_frame     *   ref_frame;
    char                *   filename;
    const char          *   sval;
    int                     nima, file_id;
    int                     nraw;

    /* Initialise */
    nima = cpl_imagelist_get_size(flat);
    if (nima != 2 && nima != 3) return CPL_ERROR_UNSPECIFIED;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();
   
    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Test the status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(qclist);
        cpl_propertylist_delete(plist);
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Add FILTER */
    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);
    /* Add OBJECTIVE */
    sval = isaac_pfits_get_objective(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC OBJECTIVE", sval);
    cpl_propertylist_delete(plist);
    cpl_propertylist_append_double(qclist, "ESO QC TWFLAT MEDMIN",
            isaac_img_twflat_config.med_min);
    cpl_propertylist_append_double(qclist, "ESO QC TWFLAT MEDMAX",
            isaac_img_twflat_config.med_max);
    cpl_propertylist_append_double(qclist, "ESO QC TWFLAT MEDAVG",
            isaac_img_twflat_config.med_avg);
    cpl_propertylist_append_double(qclist, "ESO QC TWFLAT MEDSTDEV",
            isaac_img_twflat_config.med_stdev);
    if (isaac_img_twflat_config.nb_badpix >=0 ) {
        cpl_propertylist_append_int(qclist, "ESO QC TWFLAT NBADPIX",
                isaac_img_twflat_config.nb_badpix);
    }
    cpl_propertylist_append_int(qclist, "ESO QC TWFLAT NFRAMES",
            nframes);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN LLMAX",
            isaac_img_twflat_config.oddeven_ll_max);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN LLMEAN",
            isaac_img_twflat_config.oddeven_ll_mean);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN LLSTDEV",
            isaac_img_twflat_config.oddeven_ll_stdev);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN LRMAX",
            isaac_img_twflat_config.oddeven_lr_max);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN LRMEAN",
            isaac_img_twflat_config.oddeven_lr_mean);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN LRSTDEV",
            isaac_img_twflat_config.oddeven_lr_stdev);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN ULMAX",
            isaac_img_twflat_config.oddeven_ul_max);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN ULMEAN",
            isaac_img_twflat_config.oddeven_ul_mean);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN ULSTDEV",
            isaac_img_twflat_config.oddeven_ul_stdev);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN URMAX",
            isaac_img_twflat_config.oddeven_ur_max);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN URMEAN",
            isaac_img_twflat_config.oddeven_ur_mean);
    cpl_propertylist_append_double(qclist, "ESO QC ODDEVEN URSTDEV",
            isaac_img_twflat_config.oddeven_ur_stdev);

    /* Write the flat image */
    filename = cpl_sprintf("isaac_img_twflat_set%02d.fits", set_nb);
    irplib_dfs_save_image(set_tot,
            parlist,
            set,
            cpl_imagelist_get(flat, 0),
            CPL_BPP_IEEE_FLOAT,
            "isaac_img_twflat",
            ISAAC_IMG_TWFLAT_RES,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            filename);
    cpl_free(filename);

    /* Write the bpm image */
    if (bpm) {
        filename = cpl_sprintf("isaac_img_twflat_set%02d_bpm.fits", set_nb);
        irplib_dfs_save_image(set_tot,
                parlist,
                set,
                bpm,
                CPL_BPP_IEEE_FLOAT,
                "isaac_img_twflat",
                ISAAC_IMG_TWFLAT_BPM,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                filename);
        cpl_free(filename);
    }
    
    /* Write the intercept image */
    if (isaac_img_twflat_config.intercept_flag && nima == 3) {
        filename = cpl_sprintf("isaac_img_twflat_set%02d_inter.fits",set_nb);
        irplib_dfs_save_image(set_tot,
                parlist,
                set,
                cpl_imagelist_get(flat, 1),
                CPL_BPP_IEEE_FLOAT,
                "isaac_img_twflat",
                ISAAC_IMG_TWFLAT_INTERC,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                filename);
        cpl_free(filename);
    }
    
    /* Write the error map image */
    if (isaac_img_twflat_config.errmap_flag) {
        if (nima == 2)  file_id = 1;
        else            file_id = 2;
        filename=cpl_sprintf("isaac_img_twflat_set%02d_errmap.fits", set_nb);
        irplib_dfs_save_image(set_tot,
                parlist,
                set,
                cpl_imagelist_get(flat, file_id),
                CPL_BPP_IEEE_FLOAT,
                "isaac_img_twflat",
                ISAAC_IMG_TWFLAT_ERR,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                filename);
        cpl_free(filename);
    }

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(DATE-OBS|ESO DET CHIP NAME|ARCFILE|ESO TPL ID|ESO DET MODE NAME|"
        "ESO DET NCORRS NAME|ESO DET RSPEED|ESO DET DIT)$", 0);
    cpl_propertylist_delete(plist);


    /* Add PRO.DATANCOM keyword (DFS05421) */
    nraw = 0;
    for (cpl_size i = 0; i < nframes; i++) {
        const cpl_frame* iframe = cpl_frameset_get_position_const(set, i);
        if (cpl_frame_get_group(iframe) == CPL_FRAME_GROUP_RAW) nraw++;
    }

    if (!cpl_propertylist_has(paflist, PRO_DATANCOM)) {
        cpl_propertylist_update_int(paflist, PRO_DATANCOM, nraw);
        cpl_propertylist_set_comment(paflist, PRO_DATANCOM,
                                     "Number of combined frames");
    }
    /* The above piece of code regarding DFS05421 has been partially copied
       from cpl_dfs.c as a workaround for that request until PAF files are
       completely deprecated (which is planned for the long term). */

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_TWFLAT_RES);

    /* Save the PAF file */
    filename = cpl_sprintf("isaac_img_twflat_set%02d.paf", set_nb);
    cpl_dfs_save_paf("ISAAC",
            "isaac_img_twflat",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);
    return  0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Comparison function to identify different settings
  @param    frame1  first frame 
  @param    frame2  second frame 
  @return   0 if frame1!=frame2, 1 if frame1==frame2, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_twflat_compare(
        const cpl_frame *   frame1, 
        const cpl_frame *   frame2) 
{
    int                     comparison;
    cpl_propertylist    *   plist1;
    cpl_propertylist    *   plist2;
    const char          *   sval1,
                        *   sval2;

    /* Test entries */
    if (frame1==NULL || frame2==NULL) return CPL_ERROR_UNSPECIFIED;

    /* Get property lists */
    if ((plist1=cpl_propertylist_load(cpl_frame_get_filename(frame1),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        return CPL_ERROR_UNSPECIFIED;
    }
    if ((plist2=cpl_propertylist_load(cpl_frame_get_filename(frame2),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(plist1);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Test status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }

    comparison = 1;

    /* Compare filters */
    sval1 = isaac_pfits_get_filter(plist1);
    sval2 = isaac_pfits_get_filter(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the filter");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (strcmp(sval1, sval2)) comparison = 0;

    cpl_propertylist_delete(plist1);
    cpl_propertylist_delete(plist2);
    return comparison;
}


