/* $Id: isaac_img_zpoint.c,v 1.61 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.61 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_calib.h"
#include "irplib_strehl.h"
#include "irplib_stdstar.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

#include <string.h>
#include <math.h>
#include <assert.h>

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "isaac_img_zpoint"

#define PHOT_STAR_RADIUS                30.0
#define PHOT_BACKGROUND_R1              40.0
#define PHOT_BACKGROUND_R2              60.0

#define DEF_LOCATE_SX                   10
#define DEF_LOCATE_SY                   10

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_table * isaac_img_zpoint_reduce(cpl_frameset *, const char *,
        const char *, const char *, const char *, const char *, cpl_image **);
static cpl_imagelist * isaac_img_zpoint_load(cpl_frameset *, const char *, 
        const char *, const char *, const char *);
static int isaac_img_zpoint_save(cpl_table *, cpl_image *, cpl_frameset *, 
        const cpl_parameterlist *, cpl_frameset *);
static cpl_table * isaac_img_zpoint_photom(cpl_imagelist *, cpl_bivector *);
static cpl_error_code isaac_img_zpoint_get_mag(const char *, double, double,
                                               isaac_band);
static int isaac_img_zpoint_gradients(cpl_vector *, cpl_bivector *, double *, 
        double *, double *, double *);
static cpl_bivector * isaac_img_zpoint_get_offsets(cpl_frameset *);
static cpl_error_code isaac_img_zpoint_compute_keywords(const cpl_frameset *,
                                                        double *,double *);
static cpl_image * isaac_img_zpoint_check_im(cpl_imagelist *,
        cpl_bivector *, double, double, double);

cpl_recipe_define(isaac_img_zpoint, ISAAC_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2008", 
                  "ISAAC Zero point computation recipe",
                  RECIPE_STRING " -- ISAAC Zero point recipe\n"
                  "The files listed in the Set Of Frames (sof-file) must "
                  "be tagged:\n"
                  "raw-file.fits "ISAAC_IMG_ZPOINT_RAW" or\n"
                  "raw-file.fits "ISAAC_IMG_ZPOINT_CHOPPING_RAW" or\n"
                  "stdstars.fits "ISAAC_CALIB_STDSTARS" or\n"
                  "flat-file.fits "ISAAC_CALIB_FLAT" or\n"
                  "detlin-a-file.fits "ISAAC_CALIB_DETLIN_A" or\n"
                  "detlin-b-file.fits "ISAAC_CALIB_DETLIN_B" or\n"
                  "detlin-c-file.fits "ISAAC_CALIB_DETLIN_C"\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    double      ra;
    double      dec;
    double      magnitude;
    int         sx;
    int         sy;
    double      phot_star_radius;
    double      phot_bg_r1;
    double      phot_bg_r2;
    int         chopping;
    int         check_im;
    /* Outputs */
    double      dit;
    char        filter[512];
    isaac_band  band;
    char        * starname;
    char        * sptype;
    char        * catalog;
    double      zpoint;
    double      zpointrms;
    double      flux_med;
    double      gradx;
    double      grady;
    double      graddx;
    double      graddy;
    double      fwhm_mean;
} isaac_img_zpoint_config;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_img_zpoint_fill_parameterlist(cpl_parameterlist * self)
{

    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */


    /* --star_r */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "star_r", PHOT_STAR_RADIUS, NULL,
                                          context, "The star radius");
    cpl_ensure_code(!err, err);

    /* --bg_r1 */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "bg_r1", PHOT_BACKGROUND_R1, NULL,
                                          context, "The internal background "
                                          "radius");
    cpl_ensure_code(!err, err);

    /* --bg_r2 */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "bg_r2", PHOT_BACKGROUND_R2, NULL,
                                          context, "The external background "
                                          "radius");
    cpl_ensure_code(!err, err);

    /* --ra */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "ra", 999.0, NULL, context,
                                          "Right Ascension [degrees]");
    cpl_ensure_code(!err, err);

    /* --dec */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "dec", 999.0, NULL, context,
                                          "Declination [degrees]");
    cpl_ensure_code(!err, err);

    /* --mag */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "mag", 99.0, NULL, context,
                                          "Standard Star Magnitude to use "
                                          "instead of catalogue value");
    cpl_ensure_code(!err, err);

    /* --sx */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sx", DEF_LOCATE_SX, NULL, context,
                                       "X-size of the search window");
    cpl_ensure_code(!err, err);

    /* --sy */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sy", DEF_LOCATE_SY, NULL, context,
                                       "Y-size of the search window");
    cpl_ensure_code(!err, err);

    /* --check_im */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "check_im", CPL_FALSE, NULL, context,
                                        "Flag to create the check image");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint(cpl_frameset            * framelist,
                            const cpl_parameterlist * parlist)
{

    const char   * stdstars;
    const char   * flat;
    const char   * detlin_a;
    const char   * detlin_b;
    const char   * detlin_c;
    cpl_frameset * rawframes = NULL;
    cpl_table    * tab       = NULL;
    cpl_image    * check_im  = NULL;
    
    /* Initialise */
    isaac_img_zpoint_config.starname = NULL;
    isaac_img_zpoint_config.sptype = NULL;
    isaac_img_zpoint_config.filter[0] = (char)0;
    isaac_img_zpoint_config.catalog = NULL;
    isaac_img_zpoint_config.gradx = -1.0;
    isaac_img_zpoint_config.grady = -1.0;
    isaac_img_zpoint_config.graddx = -1.0;
    isaac_img_zpoint_config.graddy = -1.0;
    isaac_img_zpoint_config.fwhm_mean = -1.0;

    /* Retrieve input parameters */
    /* --ra */
    isaac_img_zpoint_config.ra =
        irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING, "ra");

    /* --dec */
    isaac_img_zpoint_config.dec =
        irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING, "dec");

    /* --mag */
    isaac_img_zpoint_config.magnitude =
        irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING, "mag");

    /* --sx */
    isaac_img_zpoint_config.sx =
        irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING, "sx");

    /* --sy */
    isaac_img_zpoint_config.sy =
        irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING, "sy");

    /* --star_r */
    isaac_img_zpoint_config.phot_star_radius =
        irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                        "star_r");
    /* --bg_r1 */
    isaac_img_zpoint_config.phot_bg_r1 =
        irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                        "bg_r1");

    /* --bg_r2 */
    isaac_img_zpoint_config.phot_bg_r2 =
        irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                        "bg_r2");

    /* --check_im */
    isaac_img_zpoint_config.check_im =
        irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                      "check_im");

    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (isaac_dfs_set_groups(framelist));

    /* Retrieve calibration data */
    detlin_a = isaac_extract_filename(framelist, ISAAC_CALIB_DETLIN_A);
    detlin_b = isaac_extract_filename(framelist, ISAAC_CALIB_DETLIN_B);
    detlin_c = isaac_extract_filename(framelist, ISAAC_CALIB_DETLIN_C);
    flat     = isaac_extract_filename(framelist, ISAAC_CALIB_FLAT);
    stdstars = isaac_extract_filename(framelist, ISAAC_CALIB_STDSTARS);

    /* Retrieve raw frames */
    if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_ZPOINT_RAW)) != NULL) {
        isaac_img_zpoint_config.chopping = 0;
    } else if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_ZPOINT_CHOPPING_RAW)) != NULL) {
        isaac_img_zpoint_config.chopping = 1;
    } else {
        error_if(1, CPL_ERROR_DATA_NOT_FOUND, "Cannot find raw frames in the "
                 "input list");
    }

    /* Compute the zpoint values */
    cpl_msg_info(cpl_func, "Reduce the data");
    cpl_msg_indent_more();
    tab = isaac_img_zpoint_reduce(rawframes, stdstars, flat, detlin_a, 
                                  detlin_b, detlin_c, &check_im);
    cpl_msg_indent_less();
    skip_if(tab == NULL);

    /* Save the products */
    cpl_msg_info(cpl_func, "Save the paf file");
    cpl_msg_indent_more();
    isaac_img_zpoint_save(tab, check_im, rawframes, parlist, framelist);
    cpl_msg_indent_less();
    skip_if (0);

    end_skip;

    cpl_image_delete(check_im);
    cpl_frameset_delete(rawframes);
    cpl_table_delete(tab);
    cpl_free(isaac_img_zpoint_config.starname);
    cpl_free(isaac_img_zpoint_config.sptype);
    cpl_free(isaac_img_zpoint_config.catalog);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    set         the set of frames
  @param    stdstars    the standard stars catalog or NULL
  @param    flat    the flat field name or NULL 
  @param    detlin_a    the a coeffs for linearity or NULL
  @param    detlin_b    the b coeffs for linearity or NULL
  @param    detlin_c    the c coeffs for linearity or NULL
  @param    check_im    the produced check image or NULL
  @return   a cpl_table with the results or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_img_zpoint_reduce(
        cpl_frameset    *   set,
        const char      *   stdstars,
        const char      *   flat,
        const char      *   detlin_a,
        const char      *   detlin_b,
        const char      *   detlin_c,
        cpl_image       **  check_im)
{
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    const char          *   sval;
    cpl_imagelist       *   imlist;
    cpl_image           *   ima;
    cpl_mask            *   kernel;
    int                     niter;
    int                     size_x, size_y;
    double                  pos_x, pos_y, pos_x_cen, pos_y_cen, dist, min_dist;
    cpl_apertures       *   aperts;
    cpl_bivector        *   offsets,
                        *   positions;
    double                  off_x, off_y;
    int                     llx, lly, urx, ury;
    double                  val, sqsum, zprms, avg_zp;
    cpl_table           *   tab_res;
    cpl_vector          *   tmp_vec;
    int                     nb_ok;
    int                     i;

    /* Initialise */
    *check_im = NULL;
    pos_x_cen = pos_y_cen = -1.0;

    /* Get the filter name, DIT, RA and DEC  */
    cur_frame = cpl_frameset_get_position(set, 0);
    plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
    if ((sval = isaac_pfits_get_filter(plist)) == NULL) {
        cpl_propertylist_delete(plist);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }
    (void)strncpy(isaac_img_zpoint_config.filter, sval, 512);
    isaac_img_zpoint_config.filter[511] = (char)0;
    if (isaac_img_zpoint_config.ra > 998.0) 
        isaac_img_zpoint_config.ra = isaac_pfits_get_ra(plist);
    if (isaac_img_zpoint_config.dec > 998.0) 
        isaac_img_zpoint_config.dec = isaac_pfits_get_dec(plist);
    isaac_img_zpoint_config.dit = isaac_pfits_get_dit(plist);
    cpl_propertylist_delete(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot get some header informations");
        return NULL;
    }
    cpl_msg_info(cpl_func, "Using star at position: RA = %g; DEC = %g", 
            isaac_img_zpoint_config.ra, isaac_img_zpoint_config.dec);
   
    /* Get the band */
    if ((isaac_img_zpoint_config.band = 
                isaac_get_bbfilter(isaac_img_zpoint_config.filter)) ==
            ISAAC_BAND_UNKNOWN) {
        cpl_msg_error(cpl_func, "Cannot associate the filter to a BB one");
        return NULL;
    }
    
    /* Get the standard star information from database */
    if (isaac_img_zpoint_config.magnitude > 98.0) {
        cpl_msg_info(cpl_func, "Get the star magnitude");
        if (isaac_img_zpoint_get_mag(stdstars, isaac_img_zpoint_config.ra,
                                    isaac_img_zpoint_config.dec,
                                    isaac_img_zpoint_config.band)) {
            cpl_msg_error(cpl_func,"Cannot get the magnitude from the catalog");
            return NULL;
        }
    }
    cpl_msg_info(cpl_func, "Star magnitude with filter %s : %g",
            isaac_img_zpoint_config.filter,
            isaac_img_zpoint_config.magnitude);
    
    /* Load the images */
    cpl_msg_info(cpl_func, "Load images");
    if ((imlist = isaac_img_zpoint_load(set, flat, detlin_a, detlin_b,
                    detlin_c)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load the images"); 
        return NULL;
    }
       
    /* Detect the central object in the first frame */
    cpl_msg_info(cpl_func, "Detect a bright object in the first frame");
    size_x = cpl_image_get_size_x(cpl_imagelist_get(imlist, 0));
    size_y = cpl_image_get_size_y(cpl_imagelist_get(imlist, 0));
    kernel = cpl_mask_new(3, 3);
    cpl_mask_not(kernel);
    ima = cpl_image_new(size_x, size_y,
                        cpl_image_get_type(cpl_imagelist_get(imlist, 0)));
    cpl_image_filter_mask(ima, cpl_imagelist_get(imlist, 0), kernel,
                          CPL_FILTER_MEDIAN, CPL_BORDER_FILTER);
    cpl_mask_delete(kernel);
    aperts = cpl_apertures_extract_sigma(ima, 5.0);
    cpl_image_delete(ima);
    min_dist = size_x * size_x + size_y * size_y;
    for (i=0; i<cpl_apertures_get_size(aperts); i++) {
        pos_x = cpl_apertures_get_centroid_x(aperts, i+1);
        pos_y = cpl_apertures_get_centroid_y(aperts, i+1);
        dist = (pos_x-size_x/2)*(pos_x-size_x/2) +
               (pos_y-size_y/2)*(pos_y-size_y/2);
        if (dist<min_dist) {
            min_dist = dist;
            pos_x_cen = pos_x;
            pos_y_cen = pos_y;
        }
    }
    cpl_apertures_delete(aperts);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot find the central object");
        cpl_imagelist_delete(imlist);
        return NULL;
    }
    cpl_msg_info(cpl_func, "Bright object position: %g %g", 
            pos_x_cen, pos_y_cen);
   
    /* Get the offsets */
    cpl_msg_info(cpl_func, "Read the offsets in the header");
    offsets = isaac_img_zpoint_get_offsets(set);
    
    /* Deduce the objects positions  */
    niter = cpl_imagelist_get_size(imlist);
    positions   = cpl_bivector_new(niter);
    for (i=0; i<niter; i++) {
        off_x = cpl_vector_get(cpl_bivector_get_x(offsets), i) -
            cpl_vector_get(cpl_bivector_get_x(offsets), 0);
        off_y = cpl_vector_get(cpl_bivector_get_y(offsets), i) -
            cpl_vector_get(cpl_bivector_get_y(offsets), 0);
        pos_x = pos_x_cen + off_x;
        pos_y = pos_y_cen + off_y;
        cpl_vector_set(cpl_bivector_get_x(positions), i, pos_x);
        cpl_vector_set(cpl_bivector_get_y(positions), i, pos_y);
    }
    cpl_bivector_delete(offsets);

    /* Refine the positions */
    cpl_msg_info(cpl_func, "Refine the star positions");
    for (i=1; i<niter; i++) {
        pos_x = cpl_vector_get(cpl_bivector_get_x(positions), i);
        pos_y = cpl_vector_get(cpl_bivector_get_y(positions), i);
        llx = pos_x - isaac_img_zpoint_config.sx;
        urx = pos_x + isaac_img_zpoint_config.sx;
        lly = pos_y - isaac_img_zpoint_config.sy;
        ury = pos_y + isaac_img_zpoint_config.sy;
        ima = cpl_imagelist_get(imlist, i);
        pos_x = cpl_image_get_centroid_x_window(ima, llx, lly, urx, ury);
        pos_y = cpl_image_get_centroid_y_window(ima, llx, lly, urx, ury);
        cpl_vector_set(cpl_bivector_get_x(positions), i, pos_x);
        cpl_vector_set(cpl_bivector_get_y(positions), i, pos_y);
    }
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot refine the positions");
        cpl_imagelist_delete(imlist);
        cpl_bivector_delete(positions);
        return NULL;
    }

    /* Create the check image if requested */
    if (isaac_img_zpoint_config.check_im) {
        *check_im = isaac_img_zpoint_check_im(imlist, positions, 
                isaac_img_zpoint_config.phot_star_radius, 
                isaac_img_zpoint_config.phot_bg_r1,
                isaac_img_zpoint_config.phot_bg_r2);
    }

    /* Reduce the images */
    cpl_msg_info(cpl_func, "Compute the photometry");
    cpl_msg_indent_more();
    if ((tab_res = isaac_img_zpoint_photom(imlist, positions)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot reduce");
        cpl_bivector_delete(positions);
        cpl_imagelist_delete(imlist);
        cpl_image_delete(*check_im);
        *check_im = NULL;
        cpl_msg_indent_less();
        return NULL;
    } 
    cpl_msg_indent_less();
    cpl_imagelist_delete(imlist);

    /* Compute the median of the flux */
    isaac_img_zpoint_config.flux_med =
        cpl_table_get_column_median(tab_res, "FLUX"); 

    /* Get the mean FWHM  */
    isaac_img_zpoint_config.fwhm_mean =
        cpl_table_get_column_mean(tab_res, "FWHMX");
    isaac_img_zpoint_config.fwhm_mean +=
        cpl_table_get_column_mean(tab_res, "FWHMY");
    isaac_img_zpoint_config.fwhm_mean /= 2.0;

    /* Compute the flux gradient */
    tmp_vec = cpl_vector_new(niter);
    for (i=0; i<niter; i++) {
        cpl_vector_set(tmp_vec, i, 
                cpl_table_get_double(tab_res, "FLUX", i, NULL));
    }
    isaac_img_zpoint_gradients(tmp_vec, positions,
            &(isaac_img_zpoint_config.gradx),
            &(isaac_img_zpoint_config.grady),
            &(isaac_img_zpoint_config.graddx),
            &(isaac_img_zpoint_config.graddy));
    cpl_vector_delete(tmp_vec);
    cpl_bivector_delete(positions);
    
    /* Compute the averages of the results for the zero point */
    tmp_vec = cpl_vector_new(niter);
    for (i=0; i<niter; i++) {
        cpl_vector_set(tmp_vec, i, 
                cpl_table_get_double(tab_res, "ZPOINT", i, NULL));
    }
    nb_ok = 0;
    avg_zp = sqsum = 0.0;
    if (isaac_img_zpoint_config.chopping == 0) {
        cpl_vector_sort(tmp_vec, 1);
        /* Reject highest and lowest value */
        for (i=1; i<niter-1; i++) {
            val = cpl_vector_get(tmp_vec, i);
            if (val > 0.0) {
                avg_zp += val;
                sqsum += val * val;
                nb_ok ++;
            }
        }
    } else if (isaac_img_zpoint_config.chopping == 1) {
        for (i=0; i<niter; i++) {
            val = cpl_vector_get(tmp_vec, i);
            if (val > 0.0) {
                avg_zp += val;
                sqsum += val * val;
                nb_ok ++;
            }
        }
    } else {
        cpl_msg_error(cpl_func, "unsupported mode"); 
        cpl_table_delete(tab_res);
        cpl_vector_delete(tmp_vec);
        cpl_image_delete(*check_im);
        *check_im = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_vector_delete(tmp_vec);
    if (nb_ok < 1) {
        cpl_msg_error(cpl_func, "no valid zpoint measurement: cannot compute"); 
        isaac_img_zpoint_config.zpoint    = -1.0;
        isaac_img_zpoint_config.zpointrms = -1.0;
    } else {
        avg_zp /= (double)nb_ok;
        sqsum  /= (double)nb_ok;
        zprms = sqsum - avg_zp * avg_zp;
        zprms = zprms > 0 ? sqrt(zprms) : 0;
        isaac_img_zpoint_config.zpoint    = avg_zp;
        isaac_img_zpoint_config.zpointrms = zprms;
    }

    /* Print final results */
    cpl_msg_info(cpl_func, "***** FINAL RESULTS *****");
    cpl_msg_info(cpl_func, "Zero point :   %g", isaac_img_zpoint_config.zpoint);
    cpl_msg_info(cpl_func, "Zero p. RMS:   %g", isaac_img_zpoint_config.zpointrms);
    
    return tab_res;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create an image with subimages and circles around the star
  @param    imlist      List of input images
  @param    positions   List of star positions
  @param    r1          First radius
  @param    r2          Second radius
  @param    r3          Third radius
  @return   The generated image

  The function takes sub images of the input images centered around the
  specified positions, and make a big images with the sub images next to
  each others. Circles are drawn around the first central position with
  three different radii (r1, r2, r3). 
 */
/*----------------------------------------------------------------------------*/
static cpl_image * isaac_img_zpoint_check_im(
        cpl_imagelist   *   imlist,
        cpl_bivector    *   positions,
        double              r1,
        double              r2,
        double              r3)
{
    int                     nima, in_nx, in_ny, nx, ny, box_sz, llx, lly;
    cpl_image           *   in_ima;
    float               *   pin_ima;
    cpl_image           *   out_ima;
    float               *   pout_ima;
    int                     out_pos, in_pos;
    double              *   pos_x;
    double              *   pos_y;
    double                  dist;
    int                     i, j, k;
 
    /* Check entries */
    if (imlist == NULL) return NULL;
    if (positions == NULL) return NULL;
    nima = cpl_imagelist_get_size(imlist);
    if (cpl_bivector_get_size(positions) != nima) return NULL;
    
    /* Initialise */
    in_ima = cpl_imagelist_get(imlist, 0);
    in_nx = cpl_image_get_size_x(in_ima);
    in_ny = cpl_image_get_size_y(in_ima);
    pos_x = cpl_bivector_get_x_data(positions);
    pos_y = cpl_bivector_get_y_data(positions);

    /* Create the output image */
    box_sz = 2 * (int)r3 + 1;
    nx = nima * box_sz;    
    ny = box_sz;
    out_ima = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
    pout_ima = cpl_image_get_data_float(out_ima);

    /* Loop on the input images */
    for (k=0; k<nima; k++) {
        /* Get the current input image */
        in_ima = cpl_imagelist_get(imlist, k);
        pin_ima = cpl_image_get_data_float(in_ima);

        /* Get the sub image position */
        llx = (int)(pos_x[k] - r3);
        lly = (int)(pos_y[k] - r3);
        
        for (i=0; i<box_sz; i++) {
            for (j=0; j<box_sz; j++) {
                out_pos = (box_sz * k) + i + j * nx; 
                in_pos = llx + i + (lly+j) * in_nx;
                if (in_pos >= 0 && in_pos < in_nx*in_ny)
                    pout_ima[out_pos] = pin_ima[in_pos];
            }
        }
    }

    /* Draw circles */
    for (i=0; i<box_sz; i++) {
        for (j=0; j<box_sz; j++) {
            dist = sqrt((i - (box_sz/2)) * (i - (box_sz/2)) +
                (j - (box_sz/2)) * (j - (box_sz/2)));
            if ((fabs(dist-sqrt(r1*r1))) < 0.5) pout_ima[i + j * nx] = 10000;
            if ((fabs(dist-sqrt(r2*r2))) < 0.5) pout_ima[i + j * nx] = 10000;
            if ((fabs(dist-sqrt(r3*r3))) < 0.5) pout_ima[i + j * nx] = 10000;
        }
    }

    /* Return  */
    return out_ima;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load the input images
  @param
  @return
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_img_zpoint_load(
        cpl_frameset    *   set,
        const char      *   flat,
        const char      *   detlin_a,
        const char      *   detlin_b,
        const char      *   detlin_c)
{
    cpl_imagelist       *   imlist;
    cpl_imagelist       *   imlist2;
    cpl_imagelist       *   diffs;
    cpl_image           *   ima;
    int                     i;
    
    /* Load the images */
    if (isaac_img_zpoint_config.chopping == 0) {
        if ((imlist = cpl_imagelist_load_frameset(set, CPL_TYPE_FLOAT, 1,
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the images"); 
            return NULL;
        }
    } else if (isaac_img_zpoint_config.chopping == 1) {
        if ((imlist = cpl_imagelist_load_frameset(set, CPL_TYPE_FLOAT, 1,
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the images"); 
            return NULL;
        }
        if ((imlist2 = cpl_imagelist_load_frameset(set, CPL_TYPE_FLOAT, 2,
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the images"); 
            cpl_imagelist_delete(imlist);
            return NULL;
        }
        cpl_imagelist_subtract(imlist, imlist2);
        cpl_imagelist_delete(imlist2);
    } else {
        cpl_msg_error(cpl_func, "Unsupported mode"); 
        return NULL;
    }

    /* Correct the detector linearity */
    if (detlin_a && detlin_b && detlin_c) {
        cpl_msg_info(cpl_func, "Correct for non-linearity");
        if (irplib_detlin_correct(imlist, detlin_a, detlin_b, detlin_c) == -1) {
            cpl_msg_error(cpl_func, "Cannot correct for non-linearity");
        }
    }

    /* Divide by the flatfield if one is provided */
    if (flat) {
        cpl_image * flat_im = cpl_image_load(flat, CPL_TYPE_UNSPECIFIED, 0, 0);
        cpl_error_code error = flat_im ? CPL_ERROR_NONE : cpl_error_get_code();

        cpl_msg_info(cpl_func, "Divide by the flat field");

        /* Threshold all pixels below the "mid-point" of FLT_MIN and 1, and
           all pixels above the "mid-point" of 1 and FLT_MAX. */
        error |= cpl_image_threshold(flat_im, sqrt(FLT_MIN), sqrt(FLT_MAX),
                                     1.0, 1.0);
        error |= cpl_imagelist_divide_image(imlist, flat_im);
        cpl_image_delete(flat_im);

        if (error) {
            cpl_imagelist_delete(imlist);
            (void)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                        "Flat field correction failed");
            return NULL;
        }
    }

    /* Different in chopping and non-chopping modes */
    if (isaac_img_zpoint_config.chopping == 1) {
        diffs = imlist;
        imlist = NULL;
    } else if (isaac_img_zpoint_config.chopping == 0) {
        /* Build the difference image list */
        diffs = cpl_imagelist_new();
        for (i=0; i<cpl_imagelist_get_size(imlist)-1; i++) {
            ima = cpl_image_subtract_create(
                    cpl_imagelist_get(imlist, i),
                    cpl_imagelist_get(imlist, i+1));
            cpl_imagelist_set(diffs, ima, 2*i);
            ima = cpl_image_subtract_create(
                    cpl_imagelist_get(imlist, i+1),
                    cpl_imagelist_get(imlist, i));
            cpl_imagelist_set(diffs, ima, 2*i+1);
        }
        cpl_imagelist_delete(imlist);
        if (cpl_error_get_code()) {
            cpl_msg_error(cpl_func, "Cannot build the difference images");
            cpl_imagelist_delete(diffs);
            return NULL;
        }
    } else {
        cpl_msg_error(cpl_func, "Unsupported mode - abort");
        cpl_imagelist_delete(imlist);
        return NULL;
    }
    return diffs;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Reduce one image
  @param    ilist   the input image list
  @param    pos     the objects positions
  @return   the newly allocated table with the following columns:
            POSX POSY ZPOINT FLUX BGD PEAK FWHMX FWHMY
            or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_img_zpoint_photom(
        cpl_imagelist       *   ilist,
        cpl_bivector        *   pos)
{
    cpl_table       *   tab;
    int                 nbima;
    double              r, r1, r2, mag, dit;
    cpl_image       *   ima;
    double          *   pos_x;
    double          *   pos_y;
    double              bgd, fl, zp, peak, fwhm_x, fwhm_y;
    int                 i;

    /* Test entries */
    if (ilist == NULL) return NULL;
    if (pos == NULL) return NULL;

    /* Initialise */
    nbima = cpl_imagelist_get_size(ilist);
    mag = isaac_img_zpoint_config.magnitude;
    dit = isaac_img_zpoint_config.dit;
    r = isaac_img_zpoint_config.phot_star_radius;
    r1 = isaac_img_zpoint_config.phot_bg_r1;
    r2 = isaac_img_zpoint_config.phot_bg_r2;

    /* Create the output table */
    tab = cpl_table_new(nbima);
    cpl_table_new_column(tab, "POSX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POSY", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "ZPOINT", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "FLUX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "PEAK", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "BGD", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "FWHMX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "FWHMY", CPL_TYPE_DOUBLE);

    /* Loop on the images */
    pos_x = cpl_bivector_get_x_data(pos);
    pos_y = cpl_bivector_get_y_data(pos);
    for (i=0; i<nbima; i++) {
        /* Get the current image */
        ima = cpl_imagelist_get(ilist, i);
        /* Compute the photometry */
        /* Background */
        bgd = irplib_strehl_ring_background(ima, (int)(pos_x[i]), 
                (int)(pos_y[i]), (int)r1, (int)r2, IRPLIB_BG_METHOD_MEDIAN);
        /* Flux */
        fl = irplib_strehl_disk_flux(ima, 
                (int)(pos_x[i]), (int)(pos_y[i]), (int)r, bgd); 
        /* Zero Point */
        zp = mag + 2.5 * log10(fl) - 2.5 * log10(dit);
        cpl_msg_info(cpl_func, "Zero point nb %d:   %g", i+1, zp);
        /* Peak */
        peak = cpl_image_get_max_window(ima,
                (int)(pos_x[i]-5), (int)(pos_y[i]-5), 
                (int)(pos_x[i]+5), (int)(pos_y[i]+5)); 

        /* FWHM_X / FWHM_Y */
        if (cpl_image_get_fwhm(ima, (int)(pos_x[i]), (int)(pos_y[i]),
                &fwhm_x, &fwhm_y) != CPL_ERROR_NONE) {
            cpl_msg_debug(cpl_func, "Cannot compute FWHM for image %d", i+1);
            cpl_error_reset();
        }
                
        /* Put the results in the table */
        cpl_table_set_double(tab, "POSX", i, pos_x[i]);
        cpl_table_set_double(tab, "POSY", i, pos_y[i]);
        cpl_table_set_double(tab, "ZPOINT", i, zp);
        cpl_table_set_double(tab, "FLUX", i, fl);
        cpl_table_set_double(tab, "PEAK", i, peak);
        cpl_table_set_double(tab, "BGD", i, bgd);
        cpl_table_set_double(tab, "FWHMX", i, fwhm_x);
        cpl_table_set_double(tab, "FWHMY", i, fwhm_y);
    }
    return tab;
}

/*----------------------------------------------------------------------------*/
/*
  @brief    Compute the flux gradients
  @param    flux        a vector with 8 fluxes
  @param    positions   a bivector with the 8 positions
  @param    gradx       the x gradient
  @param    grady       the y gradient
  @param    graddx      the x gradient error
  @param    graddy      the y gradient error
  @return   0 if ok, -1 otherwise
     3    2
       1
     4    5
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_gradients(
        cpl_vector      *   flux,
        cpl_bivector    *   positions,
        double          *   gradx,
        double          *   grady,
        double          *   graddx,
        double          *   graddy) 
{
    int                 nflux;
    double              two, three, four, five, mean;
    double              two_pos_x, three_pos_x, four_pos_x, five_pos_x,
                        two_pos_y, three_pos_y, four_pos_y, five_pos_y;
    double          *   pflux;
    double          *   ppos_x;
    double          *   ppos_y;

    /* Test entries */
    if (flux == NULL) return CPL_ERROR_UNSPECIFIED;

    /* Initialise */
    nflux = cpl_vector_get_size(flux);
    if (nflux != 8) {
        *gradx = -1.0;
        *grady = -1.0;
        *graddx = -1.0;
        *graddy = -1.0;
        return 0;
    }
    mean = cpl_vector_get_mean(flux);

    /* Get the 4 positions fluxes */
    pflux = cpl_vector_get_data(flux);
    ppos_x = cpl_bivector_get_x_data(positions);
    ppos_y = cpl_bivector_get_y_data(positions);
    /* Flux */
    two = (pflux[1] + pflux[2]) / 2.0; 
    three = (pflux[3] + pflux[4]) / 2.0; 
    four = (pflux[5] + pflux[6]) / 2.0; 
    five = pflux[7];
    /* Positions */
    two_pos_x = (ppos_x[1] + ppos_x[2]) / 2.0;
    two_pos_y = (ppos_y[1] + ppos_y[2]) / 2.0;
    three_pos_x = (ppos_x[3] + ppos_x[4]) / 2.0; 
    three_pos_y = (ppos_y[3] + ppos_y[4]) / 2.0; 
    four_pos_x = (ppos_x[5] + ppos_x[6]) / 2.0; 
    four_pos_y = (ppos_y[5] + ppos_y[6]) / 2.0; 
    five_pos_x = ppos_x[7];
    five_pos_y = ppos_y[7];

    /* Compute grad */
    *gradx = (two + five - three - four) / 2;
    *grady = (three + two - four - five) / 2;

    /* Normalise to the nominal distance */
    *gradx *= (512 / ((two_pos_x + five_pos_x - three_pos_x - four_pos_x)/2));
    *grady *= (512 / ((three_pos_y + two_pos_y - four_pos_y - five_pos_y)/2));

    /* Compute error */
    *graddx = sqrt((two-five)*(two-five) + (three-four)*(three-four))/2.0;
    *graddy = sqrt((three-two)*(three-two) + (four-five)*(four-five))/2.0;
    
    /* Normalise */
    *gradx /= mean;
    *grady /= mean;
    *graddx /= mean;
    *graddy /= mean;

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the offsets from the headers
  @param    set     the input frame set
  @return   the newly allocated bivector 
 */
/*----------------------------------------------------------------------------*/
static cpl_bivector * isaac_img_zpoint_get_offsets(cpl_frameset * set) 
{
    int                     nframes;
    int                     noffsets;
    cpl_bivector        *   offs;
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    double                  off_x, off_y;
    int                     i;

    /* Test entries */
    if (set == NULL) return NULL;
    
    /* Initialize */
    nframes = cpl_frameset_get_size(set);
    
    /* Different in chopping and non-chopping modes */
    if (isaac_img_zpoint_config.chopping == 1) {
        offs = isaac_get_offsets(set);
    } else if (isaac_img_zpoint_config.chopping == 0) {
        noffsets = 2 * (nframes-1);

        /* Check error code */
        if (cpl_error_get_code()) return NULL;
        
        /* Create the bivector */
        offs = cpl_bivector_new(noffsets);
        
        /* Fill it */
        for (i=0; i<nframes; i++) {
            cur_frame = cpl_frameset_get_position(set, i);
            plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame),0);
            off_x = isaac_pfits_get_cumoffsetx(plist);
            off_y = isaac_pfits_get_cumoffsety(plist);
            cpl_propertylist_delete(plist);
            if (i==0) {
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i, off_y);
            } else if (i==nframes-1) {
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i-1, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i-1, off_y);
            } else {
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i, off_y);
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i-1, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i-1, off_y);
            }
        }
    } else {
        cpl_msg_error(cpl_func, "Unsupported mode - abort");
        return NULL;
    }

    /* Check if it went ok */
    if (cpl_error_get_code()) {
        cpl_bivector_delete(offs);
        cpl_msg_error(cpl_func, "Cannot find offsets in headers");
        return NULL;
    }
    return offs;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the standard star magnitude
  @param    stdstars    the standard stars catalog
  @param    frame       the input frame
  @param    mag         the computed magnitude
  @return   CPL_ERROR_NONE if ok, or else a CPL error code
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code isaac_img_zpoint_get_mag(const char * stdstars,
                                               double       ra,
                                               double       dec,
                                               isaac_band   band)
{
    cpl_errorstate prestate = cpl_errorstate_get();

    bug_if(0);
    bug_if(stdstars == NULL);

    switch (band) {
        /* SW mode */
        case ISAAC_BAND_J:
        case ISAAC_BAND_H:
        case ISAAC_BAND_K:
    case ISAAC_BAND_KS:
            if (!irplib_stdstar_find_star(stdstars, ra, dec,
                                          isaac_std_band_name(band),
                                          "LCO-Palomar.txt",
                                          &isaac_img_zpoint_config.magnitude,
                                          &isaac_img_zpoint_config.starname,
                                          &isaac_img_zpoint_config.sptype,
                                          &isaac_img_zpoint_config.catalog,
                                          NULL, NULL, 2.0) ||
                !irplib_stdstar_find_star(stdstars, ra, dec,
                                          isaac_std_band_name(band),
                                          "LCO-Palomar-NICMOS-Red-Stars.txt",
                                          &isaac_img_zpoint_config.magnitude,
                                          &isaac_img_zpoint_config.starname,
                                          &isaac_img_zpoint_config.sptype,
                                          &isaac_img_zpoint_config.catalog,
                                          NULL, NULL, 2.0) ||
                !irplib_stdstar_find_star(stdstars, ra, dec,
                                          isaac_std_band_name(band),
                                          "all",
                                          &isaac_img_zpoint_config.magnitude,
                                          &isaac_img_zpoint_config.starname,
                                          &isaac_img_zpoint_config.sptype,
                                          &isaac_img_zpoint_config.catalog,
                                          NULL, NULL, 2.0) ||
                /* Special case: swap K and Ks if needed */
                (band == ISAAC_BAND_K &&
                 !irplib_stdstar_find_star(stdstars, ra, dec,
                                           isaac_std_band_name(ISAAC_BAND_KS),
                                           "all",
                                           &isaac_img_zpoint_config.magnitude,
                                           &isaac_img_zpoint_config.starname,
                                           &isaac_img_zpoint_config.sptype,
                                           &isaac_img_zpoint_config.catalog,
                                           NULL, NULL, 2.0)) ||
                (band == ISAAC_BAND_KS &&
                 !irplib_stdstar_find_star(stdstars, ra, dec,
                                           isaac_std_band_name(ISAAC_BAND_K),
                                           "all",
                                           &isaac_img_zpoint_config.magnitude,
                                           &isaac_img_zpoint_config.starname,
                                           &isaac_img_zpoint_config.sptype,
                                           &isaac_img_zpoint_config.catalog,
                                           NULL, NULL, 2.0))) {
                cpl_errorstate_set(prestate);
            } else {
                skip_if(1);
            }
            break;
        /* LW mode  */
        case ISAAC_BAND_L:
        case ISAAC_BAND_M:
            if (!irplib_stdstar_find_star(stdstars, ra, dec,
                                          isaac_std_band_name(band),
                                          "ESO-VanDerBliek.txt",
                                          &isaac_img_zpoint_config.magnitude,
                                          &isaac_img_zpoint_config.starname,
                                          &isaac_img_zpoint_config.sptype,
                                          &isaac_img_zpoint_config.catalog,
                                          NULL, NULL, 2.0) ||
                !irplib_stdstar_find_star(stdstars, ra, dec,
                                          isaac_std_band_name(band),
                                          "MSSSO-Photometric.txt",
                                          &isaac_img_zpoint_config.magnitude,
                                          &isaac_img_zpoint_config.starname,
                                          &isaac_img_zpoint_config.sptype,
                                          &isaac_img_zpoint_config.catalog,
                                          NULL, NULL, 2.0) ||
                !irplib_stdstar_find_star(stdstars, ra, dec,
                                          isaac_std_band_name(band),
                                          "MSSSO-Spectroscopic.txt",
                                          &isaac_img_zpoint_config.magnitude,
                                          &isaac_img_zpoint_config.starname,
                                          &isaac_img_zpoint_config.sptype,
                                          &isaac_img_zpoint_config.catalog,
                                          NULL, NULL, 2.0) ||
                !irplib_stdstar_find_star(stdstars, ra, dec,
                                          isaac_std_band_name(band), "all",
                                          &isaac_img_zpoint_config.magnitude,
                                          &isaac_img_zpoint_config.starname,
                                          &isaac_img_zpoint_config.sptype,
                                          &isaac_img_zpoint_config.catalog,
                                          NULL, NULL, 2.0)) {
                cpl_errorstate_set(prestate);
            } else {
                skip_if(1);
            }
            break;
        default:
            cpl_msg_error(cpl_func, "cannot determine associated filter");
            skip_if(1);
    }

    end_skip;

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    Save the recipe product on disk
  @param    tab         the table to save
  @param    check_im    the check image or NULL
  @param    raw         the raw frames
  @param    parlist     the parameter list
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */ 
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_save(
        cpl_table           *   tab,
        cpl_image           *   check_im,
        cpl_frameset        *   raw,
        const cpl_parameterlist * parlist,
        cpl_frameset        *   set)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   paflist;
    cpl_propertylist    *   qclist;
    const cpl_frame           *   ref_frame;
    const char          *   sval;
    double                  hum, airm, extinction;

    /* Initialise */
    airm = 0;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Test the status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(qclist);
        cpl_propertylist_delete(plist);
        return CPL_ERROR_UNSPECIFIED;
    }
    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);
    cpl_propertylist_append_string(qclist, "ESO QC FILTER REF",
            isaac_std_band_name(isaac_img_zpoint_config.band));
    if (isaac_img_zpoint_compute_keywords(raw, &hum, &airm)) {
        cpl_propertylist_append_double(qclist, "ESO QC AMBI RHUM AVG", hum);
        cpl_propertylist_append_double(qclist, "ESO QC AIRMASS", airm);
    }
    cpl_propertylist_delete(plist);
    /* Get extinction */
    switch (isaac_img_zpoint_config.band) {
        case ISAAC_BAND_J:    extinction = 0.09; break;
        case ISAAC_BAND_JS:   extinction = 0.05; break;
        case ISAAC_BAND_H:    extinction = 0.04; break;
        case ISAAC_BAND_KS:   extinction = 0.06; break;
        default:        extinction = 0.00; break;
    }
    cpl_propertylist_append_double(qclist, "ESO QC ZPOINT",
            isaac_img_zpoint_config.zpoint);
    cpl_propertylist_append_double(qclist, "ESO QC ZPOINT ATX0",
            isaac_img_zpoint_config.zpoint + airm * extinction);
    cpl_propertylist_append_double(qclist, "ESO QC ZPOINTRMS",
            isaac_img_zpoint_config.zpointrms);
    cpl_propertylist_append_double(qclist, "ESO QC FLUX MED",
            isaac_img_zpoint_config.flux_med);
    cpl_propertylist_append_string(qclist, "ESO QC STDNAME",
                                   isaac_img_zpoint_config.starname ?
                                   isaac_img_zpoint_config.starname : "");
    cpl_propertylist_append_string(qclist, "ESO QC SPECTYPE",
                                   isaac_img_zpoint_config.sptype ?
                                   isaac_img_zpoint_config.sptype : "");
    cpl_propertylist_append_double(qclist, "ESO QC STARMAG",
            isaac_img_zpoint_config.magnitude);
    cpl_propertylist_append_string(qclist, "ESO QC CATNAME",
                                   isaac_img_zpoint_config.catalog ?
                                   isaac_img_zpoint_config.catalog : "");
    cpl_propertylist_append_double(qclist, "ESO QC GRADX",
            isaac_img_zpoint_config.gradx);
    cpl_propertylist_append_double(qclist, "ESO QC GRADY",
            isaac_img_zpoint_config.grady);
    cpl_propertylist_append_double(qclist, "ESO QC GRADDX",
            isaac_img_zpoint_config.graddx);
    cpl_propertylist_append_double(qclist, "ESO QC GRADDY",
            isaac_img_zpoint_config.graddy);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM MEAN",
            isaac_img_zpoint_config.fwhm_mean);

    /* Write the zpoint table  */
    irplib_dfs_save_table(set,
            parlist,
            set,
            tab,
            NULL,
            "isaac_img_zpoint",
            ISAAC_IMG_ZPOINT_TAB,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "isaac_img_zpoint.fits");
    /* Write the check image   */
    if (check_im) {
        irplib_dfs_save_image(set,
                parlist,
                set,
                check_im,
                CPL_BPP_IEEE_FLOAT,
                "isaac_img_zpoint",
                ISAAC_IMG_ZPOINT_CHECK,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "isaac_img_zpoint_check.fits");
    }
    
    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|ESO TPL ID|DATE-OBS|MJD-OBS|ESO INS MODE|"
        "ESO OBS ID|ESO DET DIT|ESO INS PIXSCALE|RA|DEC)$", 0);
    cpl_propertylist_delete(plist);

    /* Get the QC params in qclist and keys for paf in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, "ESO QC", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_ZPOINT_TAB);

    /* Save the PAF file */
    cpl_dfs_save_paf("ISAAC",
            "isaac_img_zpoint",
            paflist,
            "isaac_img_zpoint.paf");
    cpl_propertylist_delete(paflist);
    return  0;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    Compute the median of the values of a key in several frames
  @param    set     The input frame set
  @param    hum     On success, the computed humidity, otherwise undefined
  @param    airm    On success, the computed humidity, otherwise undefined
  @return   CPL_ERROR_NONE on sucess, otherwise the relevant CPL error code
 */ 
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_img_zpoint_compute_keywords(const cpl_frameset * set,
                                                 double             * hum,
                                                 double             * airm)
{
    const cpl_size  nframes = cpl_frameset_get_size(set);
    double          meanhum = 0.0;
    cpl_size        i;

    if (airm != NULL) *airm   = 0.0;
    if (hum  != NULL) *hum    = 0.0;

    cpl_ensure_code(set  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(hum  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(airm != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(nframes > 0,  CPL_ERROR_DATA_NOT_FOUND);

    for (i = 0; i < nframes; i++) {
        const cpl_frame  * cur_frame = cpl_frameset_get_position_const(set, i);
        const char       * filename  = cpl_frame_get_filename(cur_frame);
        cpl_propertylist * plist     = cpl_propertylist_load(filename, 0);
        const double       humi      = isaac_pfits_get_humidity_level(plist);

        if (i==0)         *airm += isaac_pfits_get_airmass_start(plist);
        if (i==nframes-1) *airm += isaac_pfits_get_airmass_end(plist);

        cpl_propertylist_delete(plist);

        meanhum += (humi - meanhum) / (double)(i + 1);

        if (cpl_error_get_code()) break;

    }

    if (cpl_error_get_code()) return cpl_error_set_where(cpl_func);

    assert(i == nframes);

    *hum   = meanhum;
    *airm *= 0.5;

    return CPL_ERROR_NONE;
}


